/*
 * console.c
 *
 *  Created on: Feb 28, 2016
 *      Author: gkubiak
 */

#include <stdio.h>
#include <string.h>
#include <stdint.h>

#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"

#include "uartRead.h"
#include "led.h"
#include "pwm.h"
#include "mpu6050.h"
#include "speed.h"
#include "esp8266.h"

#define CONSOLE_MAX_CMD_LEN     255
#define CONSOLE_MAX_NUM_ARGS    32

typedef void (*commandCallback)(uint32_t, char**);

typedef struct {
    char* cmdStr;
    uint32_t minArguments;
    commandCallback callback;
} ConsoleCommand;

typedef enum {CONSOLE_START = 0, CONSOLE_READING} ConsoleState;

static void console_setLED(uint32_t argc, char** argv);
static void console_setPWM(uint32_t argc, char** argv);
static void console_getIMU(uint32_t argc, char** argv);
static void console_monitorSpeed(uint32_t argc, char** argv);

static ConsoleCommand commands[] = {
    {"setLED", 1, console_setLED},
    {"setPWM", 1, console_setPWM},
    {"getIMU", 0, console_getIMU},
    {"monitorSpeed", 0, console_monitorSpeed},
    {NULL, 0, NULL}
};

static void writeOut(char* string)
{
	uartWrite(string);
	espWriteStringTCP(string);
}

static uint32_t createArgv(char* cmd, char** argv)
{
    uint32_t numArgs = 0;

    while (numArgs < CONSOLE_MAX_NUM_ARGS && *cmd != '\0')
    {
        if (*cmd == ',')
        {
            *cmd = '\0';
            argv[numArgs] = cmd + 1;
            numArgs++;
        }

        cmd++;
    }

    argv[numArgs] = NULL;

    return numArgs;
}

static void processCommand(char* cmd)
{
    int i = 0;
    char* argv[CONSOLE_MAX_NUM_ARGS];
    uint32_t argc;

    while (commands[i].cmdStr != NULL)
    {
        if ((strncmp(cmd, commands[i].cmdStr, strlen(commands[i].cmdStr)) == 0))
        {
            argc = createArgv(cmd, argv);

            writeOut("\r\n<");
            if (argc < commands[i].minArguments)
            {
                printf("ERROR: Need %lu Args", commands[i].minArguments);
            } else {
                commands[i].callback(argc, argv);
            }
            writeOut(">\r\n");

            break;
        }
        i++;
    }

    if (commands[i].cmdStr == NULL)
    {
        writeOut("\r\n<ERROR: Command Not Found>\r\n");
    }
}

void consoleTask( void *parameters )
{
    static ConsoleState state = CONSOLE_START;
    static char cmdBuffer[CONSOLE_MAX_CMD_LEN];
    static uint32_t cmdBufferIdx;

    uint8_t data;

    while(1)
    {
    	xQueueReceive( uartReadQueue, &data, portMAX_DELAY);

        uartWriteChar(data);
        switch(state) {
            case CONSOLE_START:
                if (data == '[')
                {
                    state = CONSOLE_READING;
                    memset(cmdBuffer, 0, CONSOLE_MAX_CMD_LEN);
                    cmdBufferIdx = 0;
                }
                break;
            case CONSOLE_READING:
                if (data == ']')
                {
                    processCommand(cmdBuffer);
                    state = CONSOLE_START;
                } else {
                    if (cmdBufferIdx < CONSOLE_MAX_CMD_LEN)
                    {
                        cmdBuffer[cmdBufferIdx] = data;
                        cmdBufferIdx++;
                    } else {
                        state = CONSOLE_START;
                    }
                }
                break;
            default:
                state = CONSOLE_START;
                break;
        }
    }
}

static void console_setLED(uint32_t argc, char** argv)
{
	writeLED(atoi(argv[0]));
}

static void console_setPWM(uint32_t argc, char** argv)
{
	setDutyCycle(atoi(argv[0]));
}

static void console_getIMU(uint32_t argc, char** argv)
{
	char buf[256];

	imuData data = getIMUData();

	sprintf(buf, "accel X: %d, Y: %d, Z: %d gyro: X: %d, Y: %d, Z: %d",
			data.accelX, data.accelY, data.accelZ,
			data.gyroX, data.gyroY, data.gyroZ);

	writeOut(buf);
}

static void console_monitorSpeed(uint32_t argc, char** argv)
{
	u32 speed;
	char buf[32];

	while (1)
	{
		speed = readSpeed();

		memset(buf, 0, 32);
		sprintf(buf, "speed: %d", speed);

		writeOut(buf);

		vTaskDelay(configTICK_RATE_HZ / 10);
	}

}
