/*
 * esp8266.c
 *
 *  Created on: Mar 3, 2016
 *      Author: gerik
 */

#include <string.h>

#include "xparameters.h"
#include "xuartps.h"
#include "xscugic.h"
#include "xil_exception.h"

#include "FreeRTOS.h"
#include "task.h"
#include "semphr.h"
#include "uartRead.h"

#define ESP8266_UART_DEVICE_ID XPAR_XUARTPS_0_DEVICE_ID
#define UART_INTERRUPT_ID XPAR_XUARTPS_0_INTR
#define ESP_RECV_BUFFER_SIZE	512

#define SERVER_TCP_IP	"172.20.10.3"
#define SERVER_TCP_PORT	"55670"

#define RESET_DELAY_TICKS (3 * configTICK_RATE_HZ)

typedef enum {ESP_INITIALIZATION = 0,
			  ESP_WAITING_FOR_HEADER,
			  ESP_READING_BYTES,
			  ESP_READING_DATA
} ESPState;

static XUartPs esp8266Uart;
extern XScuGic xInterruptController;

static const char* cmdReset = "AT+RST\r\n";
static const char* cmdSetMode = "AT+CWMODE=1\r\n";
static const char* cmdConnect = "AT+CWJAP=\"iPhone\",\"espTesting\"\r\n";
static const char* cmdInitTCP1 = "AT+CIPSTART=\"TCP\",\"";
static const char* cmdInitTCP2 = "\",";
static const char* cmdInitTCP3 = "\r\n";

static const char* cmdSendData1 = "AT+CIPSEND=";
static const char* cmdSendData2 = "\r\n";

static const char* recvDataStr = "+IPD,";

static void writeDataBlocking(char* data);
static void esp8266RecvHandler(void* callback, u32 event, unsigned int eventData);

QueueHandle_t espReadQueue = NULL;
static int espConnected = 0;

static int SetupInterruptSystem(XScuGic *IntcInstancePtr,
				XUartPs *UartInstancePtr,
				u16 UartIntrId)
{
	int Status;

	XScuGic_Config *IntcConfig; /* Config for interrupt controller */

	Xil_ExceptionRegisterHandler(XIL_EXCEPTION_ID_INT,
				(Xil_ExceptionHandler) XScuGic_InterruptHandler,
				IntcInstancePtr);

	Status = XScuGic_Connect(IntcInstancePtr, UartIntrId,
				  (Xil_ExceptionHandler) XUartPs_InterruptHandler,
				  (void *) UartInstancePtr);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XScuGic_Enable(IntcInstancePtr, UartIntrId);

	 Xil_ExceptionEnable();

	return XST_SUCCESS;
}

int initESP8266()
{
	XUartPs_Config* Config;
	XScuGic_Config* intrConfig;
	int Status;
	u32 IntrMask;

	Config = XUartPs_LookupConfig(ESP8266_UART_DEVICE_ID);
	if (NULL == Config) {
		return XST_FAILURE;
	}

	Status = XUartPs_CfgInitialize(&esp8266Uart, Config, Config->BaseAddress);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	/* Check hardware build */
	Status = XUartPs_SelfTest(&esp8266Uart);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XUartPs_SetFifoThreshold(&esp8266Uart, 1);


	/*
	 * Connect the UART to the interrupt subsystem such that interrupts
	 * can occur. This function is application specific.
	 */
	Status = SetupInterruptSystem(&xInterruptController, &esp8266Uart, UART_INTERRUPT_ID);
	if (Status != XST_SUCCESS) {
		return XST_FAILURE;
	}

	XUartPs_SetHandler(&esp8266Uart, (XUartPs_Handler)esp8266RecvHandler, &esp8266Uart);

	IntrMask = XUARTPS_IXR_RXFULL | XUARTPS_IXR_RXOVR;

	XUartPs_SetInterruptMask(&esp8266Uart, IntrMask);


	return 0;
}

static void writeDataBlocking(char* data)
{
	XUartPs_Send(&esp8266Uart, data, strlen(data));
}

static void espSendTCPString(char* string)
{
	char lenBuffer[16];

	writeDataBlocking(cmdSendData1);
	sprintf(lenBuffer, "%u", strlen(string));

	writeDataBlocking(lenBuffer);
	writeDataBlocking(cmdSendData2);

	vTaskDelay(configTICK_RATE_HZ / 100);

	writeDataBlocking(string);
	writeDataBlocking(cmdSendData2);

}

static int getEspReturnVal()
{
	char* errorString = "ERROR";
	char* okString = "OK";
	u32 errorIdx = 0;
	u32 okIdx = 0;
	u8 data;
	BaseType_t status;

	while (1)
	{
		status = xQueueReceive(espReadQueue, &data, 5 * configTICK_RATE_HZ);

		if (status == pdFALSE)
		{
			return -1;
		}

		if (data == errorString[errorIdx])
		{
			errorIdx++;
			if (errorString[errorIdx] == '\0')
			{
				return 0;
			}
		} else {
			errorIdx = 0;
		}

		if (data == okString[okIdx])
		{
			okIdx++;
			if (okString[okIdx] == '\0')
			{
				return 1;
			}
		} else {
			okIdx = 0;
		}
	}
}

void sendEspInitCmds()
{
	BaseType_t status;

	while(1) {
		vTaskDelay(RESET_DELAY_TICKS);

		uartWrite("Resting ESP\r\n");
		writeDataBlocking(cmdReset);

		vTaskDelay(RESET_DELAY_TICKS);

		xQueueReset(espReadQueue);

		uartWrite("Setting Mode\r\n");
		writeDataBlocking(cmdSetMode);
		status = getEspReturnVal();
		if (status <= 0)
		{
			continue;
		}

		uartWrite("Connecting\r\n");
		writeDataBlocking(cmdConnect);
		status = getEspReturnVal();
		if (status <= 0)
		{
			continue;
		}

		uartWrite("Connect to TCP socket\r\n");
		writeDataBlocking(cmdInitTCP1);
		writeDataBlocking(SERVER_TCP_IP);
		writeDataBlocking(cmdInitTCP2);
		writeDataBlocking(SERVER_TCP_PORT);
		writeDataBlocking(cmdInitTCP3);
		status = getEspReturnVal();
		if (status <= 0)
		{
			continue;
		}

		break;
	}
}

int waitEspForHeader()
{
	int recvIndex = 0;
	char data;
	BaseType_t status;

	while(1)
	{
		status = xQueueReceive(espReadQueue, &data, 60 * configTICK_RATE_HZ);

		if (status == pdFALSE)
		{
			return -1;
		}

		if (recvDataStr[recvIndex] == data)
		{
			recvIndex++;
			if (recvDataStr[recvIndex] == '\0')
			{
				return 1;
			}
		} else {
			recvIndex = 0;
		}
	}
}

int waitEspForBytes()
{
	char recvBuffer[8];
	int i = 0;
	char data;
	int bytes;
	BaseType_t status;

	while(1)
	{
		status = xQueueReceive(espReadQueue, &data, 3 * configTICK_RATE_HZ);

		if (status == pdFALSE)
		{
			return -1;
		}

		if (data == ':')
		{
			recvBuffer[i] = '\0';
			return atoi(recvBuffer);
		} else if (data < '0' || data > '9'){
			return -1;
		} else {
			recvBuffer[i] = data;
			i++;
		}
	}
}

int waitEspForData(int bytes, char* buffer)
{
	int i = 0;
	char data;
	BaseType_t status;

	while (1)
	{
		status = xQueueReceive(espReadQueue, &data, 3 * configTICK_RATE_HZ);

		if (status == pdFALSE)
		{
			return -1;
		}

		buffer[i] = data;
		i++;
		if (i == bytes)
		{
			return 1;
		}
	}
}

void esp8266Task(void* parameters)
{
	int status;
	ESPState state = ESP_INITIALIZATION;
	u8 data;
	u32 bytesToRead;
	u8 recvBuffer[64];
	u32 i;

	initESP8266();

	while (1)
	{
		switch (state)
		{
		case ESP_INITIALIZATION:
			espConnected = 0;
			sendEspInitCmds();
			espSendTCPString("Ready\r\n");
			state = ESP_WAITING_FOR_HEADER;
			break;
		case ESP_WAITING_FOR_HEADER:
			espConnected = 1;
			status = waitEspForHeader();
			if (status == 1)
			{
				state = ESP_READING_BYTES;
			} else {
				state = ESP_INITIALIZATION;
			}
			break;
		case ESP_READING_BYTES:
			status = waitEspForBytes();
			if (status > 0 && status < 64)
			{
				bytesToRead = status;
				state = ESP_READING_DATA;
			} else {
				state = ESP_INITIALIZATION;
			}
			break;
		case ESP_READING_DATA:
			status = waitEspForData(bytesToRead, recvBuffer);
			if (status == 1)
			{
				for (i = 0; i < bytesToRead; i++)
				{
					xQueueSend(uartReadQueue, &recvBuffer[i], 0);
				}

				state = ESP_WAITING_FOR_HEADER;
			} else {
				state = ESP_INITIALIZATION;
			}
			break;
		}
	}
}

void espWriteStringTCP(char* string)
{
	if (espConnected)
	{
		espSendTCPString(string);
	}
}

static void esp8266RecvHandler(void* callback, u32 Event, unsigned int eventData)
{
	u8 data;
	BaseType_t status;
	BaseType_t higherTaskWoken;

	if (Event == XUARTPS_EVENT_RECV_DATA)
	{
		while (XUartPs_IsReceiveData(esp8266Uart.Config.BaseAddress))
		{
			XUartPs_Recv(&esp8266Uart, &data, 1);
			status = xQueueSendFromISR(espReadQueue, &data, &higherTaskWoken);
		}
	}
}
